#!/usr/bin/perl -w

# trafgraf.pl
# a bridge between the iproute2 tool tc and snmp

# base source: http://www.docum.org/stef.coene/qos/gui/
#	but *heavily* modified 
# october 2006, *heavily* modified another time to be pass persistent  ;)
# I've used some code from dhcpd-snmp script <http://www.net-track.ch/opensource/dhcpd-snmp/>
# report bug to : Sbastien Cramatte <scramatte@zensoluciones.com>


use strict;

sub nics ();
sub disp ($$);


#------------------------------------------------------------#
my $base = "/usr/local/sbin";
my $tccmd = "/sbin/tc";
my $ipcmd = "/sbin/ip";

my $cache_secs = 60;
my $mibtime = 0;

my ($mib,$oid_in,$oid, $found);

my ($key,%bytes,%lended);

my @nics = nics() ;


#------------------------------------------------------------#
# figure out what to send back

# Switch on autoflush
$| = 1;

my $baseoid = '.1.3.6.1.4.1.2021.255';


# Main loop
while (my $cmd = <STDIN>) {
  chomp $cmd;

  if ($cmd eq "PING") {
    print "PONG\n";
  } elsif ($cmd eq "get") {
    $oid_in = <STDIN>;

    $oid = get_oid($oid_in);
    $mib = create_mib(); 
 
 		if (defined($mib->{$oid})) {
      print "$baseoid.$oid\n";
      print $mib->{$oid}[0]."\n";
      print $mib->{$oid}[1]."\n";
    } else {
      print "NONE\n";
    }    
  } elsif ($cmd eq "getnext") {
    $oid_in = <STDIN>;

    $oid = get_oid($oid_in);
    $mib = create_mib(); 
		$found = 0;

    my @s = sort { oidcmp($a, $b) } keys %{ $mib };
    for (my $i = 0; $i < @s; $i++) {
      if (oidcmp($oid, $s[$i]) == -1) {
				print "$baseoid.".$s[$i]."\n";
				print $mib->{$s[$i]}[0]."\n";
				print $mib->{$s[$i]}[1]."\n";
				$found = 1;
				last;
     	}
    }
    
    if (!$found) {
     print "NONE\n";
    }    
  } else {
  	# Unknown command
  }
 
}

#------------------------------------------------------------#
# get all the info from all interfaces (qdisc + class)

sub create_mib {
	my (%qos, %tokens, @qdisc,@class,$line,$interface, $interfaceNum, %parent, %classdatas, $number, %tmp, %oids);
	my ($byts,$pkts,$drops,$olimits,$requeues);

  if (time - $mibtime < $cache_secs) {
    return $mib;
  }

	my $numberSnmp = 0;
	for ( $interfaceNum=0; $interfaceNum < @nics; $interfaceNum++) {
		$interface = $nics[$interfaceNum];
		
		@qdisc = `$tccmd -s -d qdisc show dev $interface`;
		@class = `$tccmd -s -d class show dev $interface`;
		
		$number = "";

		#my $interfaceNumSnmp = $interfaceNum+1;
		my $interfaceNumSnmp = 2;
    		#my $numberSnmp = 0;
 
		# Don't index interfaces with no qdiscs defined
		if ($#qdisc > 1) {
			#$qos{"1.1.$interfaceNumSnmp.1"} = [ "string", $interface ];	
			#$qos{"1.2.$interfaceNumSnmp.1"} = [ "integer", $interfaceNumSnmp ];	
			#$qos{"1.2.$interfaceNumSnmp.1"} = [ "string", $#qdisc ];	
		}
		else {
			# skip ahead to next NIC, this one has no qdiscs
			next;
		}

		foreach $line (@qdisc) {
			chomp($line);
			if ($line =~ /^\s*$/) { next; }
	
			#print "$line\n";
			# get qdisc id
			# qdisc sfq 8045: quantum 1514b limit 128p flows 128/1024
			if ($line =~ /^\s*qdisc/) {
				#($number) = $line =~ /\s+([A-Za-z0-9]+):\s+/o;
				($number) = $line =~ /qdisc\s+(.*):\s+/o;
			}
	
			# Sent 0 bytes 0 pkts (dropped 0, overlimits 0)
			if (($line =~ /^\s*Sent/) && $number) {
				#($byts,$pkts,$drops,$olimits) = $line =~ /Sent\s+(\d+)\s+bytes\s+(\d+)\s+pkts\s+\(dropped\s+(\d+),\s+overlimits\s+(\d+)\)/;
				($byts,$pkts,$drops,$olimits,$requeues) = $line =~ /Sent\s+(\d+)\s+bytes\s+(\d+)\s+pkts\s+\(dropped\s+(\d+),\s+overlimits\s+(\d+)\s+requeues\s+(\d+)/i;
				#print "Matched $byts bytes $pkts pkts $drops dropped $olimits overlimits $requeues requeues\n";
				#print "Line->$line<-\n";

				$numberSnmp++;	
				#print "Calling adjust_to_32bit for $interfaceNumSnmp $numberSnmp $interface\n";					
				$qos{"2.$interfaceNumSnmp.0.$numberSnmp"} = [ "string", $interface ]; 
				$qos{"2.$interfaceNumSnmp.1.$numberSnmp"} = [ "string", $number ]; 
				$qos{"2.$interfaceNumSnmp.2.$numberSnmp"} = [ "counter", adjust_to_32bit ($byts) ]; 
				$qos{"2.$interfaceNumSnmp.3.$numberSnmp"} = [ "counter", adjust_to_32bit ($pkts) ]; 
				$qos{"2.$interfaceNumSnmp.4.$numberSnmp"} = [ "counter", adjust_to_32bit ($drops) ]; 
				$qos{"2.$interfaceNumSnmp.5.$numberSnmp"} = [ "counter", adjust_to_32bit ($olimits) ]; 
				$qos{"2.$interfaceNumSnmp.6.$numberSnmp"} = [ "counter", adjust_to_32bit ($requeues) ]; 
				
				$number = "";
				
			}
			
		}
	
		# let's skip the class info
		next;

		$number="";
		$numberSnmp=0;


		foreach $line (@class) {
			chomp ($line);
					
			if ($line =~ /^\s*$/) { next; }
	
			#print "$line\n\n";
			if ($line =~ /\s+(([A-Za-z0-9]+:)[A-Za-z0-9]+)\s+root/o) {
				$number = $1;
				$numberSnmp++;
				
				$qos{"3.$interfaceNumSnmp.1.$numberSnmp"} = [ "string", $number ]; 
				$qos{"3.$interfaceNumSnmp.2.$numberSnmp"} = [ "string", $2 ]; 
			}

			if ($line =~ /\s+([A-Za-z0-9]+:[A-Za-z0-9]+)\s+parent\s+([A-Za-z0-9]+:[A-Za-z0-9]+)\s+leaf\s+([A-Za-z0-9]+:)\s+/o) { 
				$number = $1;
				$numberSnmp++;
				
			  $qos{"3.$interfaceNumSnmp.1.$numberSnmp"} = [ "string", $number]; 
			  $qos{"3.$interfaceNumSnmp.2.$numberSnmp"} = [ "string", $2]; 
				next ;
			}

			if ($line =~ /\s+([A-Za-z0-9]+:[A-Za-z0-9]+)\s+parent\s+([A-Za-z0-9]+:[A-Za-z0-9]+)\s+/o) {
				$number = $1;
				$numberSnmp++;
				
			  $qos{"3.$interfaceNumSnmp.1.$numberSnmp"} = [ "string", $number]; 
			  $qos{"3.$interfaceNumSnmp.2.$numberSnmp"} = [ "string", $2]; 
				next ;
			}
	
			if ($line =~ /Sent\s+(\d+)\s+bytes\s+(\d+)\s+pkts\s+\(dropped\s+(\d+),\s+overlimits\s+(\d+)\)/) {
				$qos{"3.$interfaceNumSnmp.3.$numberSnmp"} = [ "counter", adjust_to_32bit ($1) ]; 
				$qos{"3.$interfaceNumSnmp.4.$numberSnmp"} = [ "counter", adjust_to_32bit ($2) ]; 
				$qos{"3.$interfaceNumSnmp.5.$numberSnmp"} = [ "counter", adjust_to_32bit ($3) ]; 
				$qos{"3.$interfaceNumSnmp.6.$numberSnmp"} = [ "counter", adjust_to_32bit ($4) ]; 
				next;
			}

			if ($line =~ /lended:\s+(\d+)\s+borrowed:\s+(\d+)\s+giants:\s+(\d+)/) {
				$qos{"3.$interfaceNumSnmp.7.$numberSnmp"} = [ "counter", adjust_to_32bit ($1) ]; 
				$qos{"3.$interfaceNumSnmp.8.$numberSnmp"} = [ "counter", adjust_to_32bit ($2) ]; 
				$qos{"3.$interfaceNumSnmp.9.$numberSnmp"} = [ "counter", adjust_to_32bit ($3) ]; 
				next ;
			}

			if ($line =~ /tokens:\s+(\d+)\s+ctokens:\s+(\d+)/) {
				$qos{"3.$interfaceNumSnmp.10.$numberSnmp"} = [ "counter", adjust_to_32bit ($1) ]; 
				$qos{"3.$interfaceNumSnmp.11.$numberSnmp"} = [ "counter", adjust_to_32bit ($2) ]; 
			  next ;
			}

		}			



	}
	
	$mib = \%qos;
  $mibtime = time;
  return $mib;
} 

sub create_coid {
	my ($k, %p, %oids) = @_;
	my ($ifoid,$val) = split /\s/, $k;

	my $id = $p{$val};
	my $oid= "";

	while (defined($id)) {
  	$oid = $oids{$id}?$oids{$id}++:1;
	  print "$oid\n";
		$id = $p{$id};
	}
	
	return "";
}


sub adjust_to_32bit {
	my ($val) = @_;
	if ($val > 4294967295) {
		$val = $val % 4294967295;
	}
	return $val;
}

sub get_oid
{

  my ($oid) = @_;
  chomp $oid;

  my $base = $baseoid;
  $base =~ s/\./\\./g;

  if ($oid !~ /^$base(\.|$)/) {
    # Requested oid doesn't match base oid
    return 0;
  }

  $oid =~ s/^$base\.?//;
  return $oid;
}


sub oidcmp {
  my ($x, $y) = @_;

  my @a = split /\./, $x;
  my @b = split /\./, $y;

  my $i = 0;

  while (1) {

    if ($i > $#a) {
      if ($i > $#b) {
	return 0;
      } else {
	return -1;
      }
    } elsif ($i > $#b) {
      return 1;
    }

    if ($a[$i] < $b[$i]) {
      return -1;
    } elsif ($a[$i] > $b[$i]) {
      return 1;
    }

    $i++;
  }
}

#------------------------------------------------------------#
# Find all configured nics
sub nics() {
	my @niclist = `$ipcmd link show`;
	my (@nic, $line);
	foreach $line (@niclist) {
		$line =~ /^([0-9]+):\s+([a-z0-9]+)/ or next;
		push(@nic, $2);
	}
	return(@nic);
}


